package pl.edu.amu.wmi.min.relativecardinality.algorithm;

import static com.google.common.math.DoubleMath.fuzzyCompare;
import java.util.HashSet;
import java.util.Set;
import org.apache.commons.math3.optim.PointValuePair;
import pl.edu.amu.wmi.min.relativecardinality.functions.Functions;

/**
 *
 * @author bikol
 */
public class GEKMAlgorithm implements Algorithm {

    public static final double K = 2.6;
    public static final double EPS = 1E-6;

    @Override
    public PointValuePair computeLowerBound(double[] wl, double[] wu, double[] x, Functions functions, Accuracy accuracy, Monitor monitor) {
        return compute(true, wl, wu, x, functions, accuracy, monitor);
    }

    @Override
    public PointValuePair computeUpperBound(double[] wl, double[] wu, double[] x, Functions functions, Accuracy accuracy, Monitor monitor) {
        return compute(false, wl, wu, x, functions, accuracy, monitor);
    }

    private PointValuePair compute(final boolean lower, final double[] wl, final double[] wu, final double[] values, final Functions functions, Accuracy accuracy, final Monitor m) {
        assert wl.length == values.length && wu.length == values.length;

        m.started();

        final Sum sum = new Sum(functions, lower, accuracy);

        final int n = values.length;
        int k = (int) ((double) n / K);

        for (int i = 0; i < n; i++) {
            sum.addElement(values[i], wl[i], wu[i]);
        }

        sum.init();
        m.afterInit();

        double nomInit = sum.getNominator();
        double denInit = sum.getDenominator();

        double baseY = sum.getY();

        for (int i = 0; i <= k; i++) {
            sum.changeValue(i, nomInit, denInit, m);
        }
        double y;
        double lastChange = -1;
        int kp = k;
        double lastY = lower ? 1 : 0;
        double error;

        Set<Integer> yValues = new HashSet<>(10);
        int iters = 0;

        do {
            y = sum.getY();
            m.iterationStarted(y, lastChange, k);

            do {
                m.innerIterationStarted(y, lastChange, k);

                k = kp;
                kp = sum.currentK();

                //-1 może oznaczać że nie można już poprawić tego wyniku
                // trzeba poprawic bo się zapętla bo optymalizuje dochodzi do k=0 i w nastepnym kroku daje kp=-1 i zeruje wszystko znów.
                //wyifować -1 albo zmianę y w złym kierunku malenie zamiast wzrostu albo na odwrót
                if (kp < 0) {
                    if (lower) {
                        if (fuzzyCompare(baseY, y, 1e-8) >= 0) {
                            kp = k;
                        }
                    } else {
                        if (fuzzyCompare(baseY, y, 1e-8) <= 0) {
                            kp = k;
                        }
                    }
                }

                //dodać numery iteracji do pliku wynikowego
                nomInit = sum.getNominator();
                denInit = sum.getDenominator();
                for (int i = 0; i <= Math.min(n - 1, kp); i++) {
                    sum.changeValue(i, nomInit, denInit, m);
                }
                for (int i = kp + 1; i <= Math.min(k, n - 1); i++) {
                    sum.resetValue(i);
                }
                lastChange = y - sum.getY();
                y = sum.getY();

                //watchdog
                //obciąc y az do odpowiedniej dokladnosci 1e-7
                yValues.add((int) (y * 10000000.0));
                iters++;
                if (yValues.size() * 4 < iters) {
                    m.solutionFound(Double.NaN, Double.NaN, -666, Double.NaN, Double.NaN);
                    m.finished();
                    System.out.println("\npetla\n");
                    return new PointValuePair(wl, Double.NaN, false);
                }
            } while (k != kp && Math.abs(lastChange) > EPS);
            error = Math.abs(lastY - y);
            lastY = y;
        } while (error > EPS);

        m.solutionFound(y, error, k, sum.getNominator(), sum.getDenominator());

        double[] point = new double[n];
        for (int i = 0; i < n; i++) {
            point[i] = sum.getValueInOriginalOrder(i);
            assert point[i] <= sum.getElementInOriginalOrder(i).wu;
            assert point[i] >= sum.getElementInOriginalOrder(i).wl;
        }

        m.finished();
        return new PointValuePair(point, y, false);
    }

}
