package pl.edu.amu.wmi.min.relativecardinality.algorithm;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import org.apache.commons.math3.util.FastMath;
import pl.edu.amu.wmi.min.relativecardinality.algorithm.Algorithm.Accuracy;
import pl.edu.amu.wmi.min.relativecardinality.algorithm.Algorithm.Monitor;
import pl.edu.amu.wmi.min.relativecardinality.functions.Functions;

/**
 *
 * @author bikol
 */
public class Sum {

    private final Functions func;
    private final List<ExtendedElement> elements = new ArrayList<>();
    private final List<ExtendedElement> elementsOrig = new ArrayList<>();
    private double nominator = -1.0, denominator = -1.0;
    private final boolean lower;
    private boolean init = false;
    private Comparator cmp;
    private final Accuracy accuracy;

    public Sum(Functions functions, boolean l, Accuracy accuracy) {
        this.lower = l;
        this.func = functions;
        this.accuracy = accuracy;
    }

    public void addElement(double x, double wl, double wu) {
        assert !init;
        assert wl <= wu;
        ExtendedElement ee = new ExtendedElement(func.p(lower, x, wl, wu), x, wl, wu);
        elements.add(ee);
        elementsOrig.add(ee);
    }

    public void init() {
        assert !init;
        nominator = -1.0;
        denominator = -1.0;
        cmp = new Comparator<Element>() {

            @Override
            public int compare(Element o1, Element o2) {
                return Double.compare(o1.p, o2.p);
            }
        };
        if (!lower) {
            cmp = Collections.reverseOrder(cmp);
        }
        Collections.sort(elements, cmp);
        init = true;
    }

    public int getN() {
        assert init;
        return elements.size();
    }

    /**
     *
     * @return -1 if no elements are such that e.p &lt; y (e.p &gt; y). n-1 if
     * all elements are such.
     */
    public int currentK() {
        assert init;
        double y = getNominator() / getDenominator();
        Element[] elems = elements.toArray(new Element[elements.size()]);
        return doCurrentK(elems, y, cmp);
    }

    static int doCurrentK(Element[] elems, double y, Comparator<Element> cmp) {
        int rv = Arrays.binarySearch(elems, new Element(y, 0, 0, 0), cmp);
        if (rv >= 0) {
            //element found, rv is position of any element with e.p==y, we need first one
//            while (rv > 0 && elems[rv - 1].p == y) {
//                rv--;
//            }
            //element found, rv is position of any element with e.p==y, we need the last one
            while (rv +1<elems.length && elems[rv + 1].p == y) {
                rv++;
            }
            return rv;
        } else {
            return -2 - rv;
        }
    }

    public void resetValue(int i) {
        assert init;
        ExtendedElement e = elements.get(i);
        nominator += func.f(e.x, e.wl) - func.f(e.x, e.w);
        denominator += func.g(e.wl) - func.g(e.w);
        e.w = e.wl;
    }

    public double getValue(int i) {
        assert init;
        return elements.get(i).w;
    }

    public double getValueInOriginalOrder(int i) {
        assert init;
        return elementsOrig.get(i).w;
    }

    public Element getElement(int i) {
        assert init;
        return elements.get(i);
    }

    public Element getElementInOriginalOrder(int i) {
        assert init;
        return elementsOrig.get(i);
    }

    public void changeValue(int i, double nom, double den, Monitor monitor) {
        assert init;
        ExtendedElement e = elements.get(i);

        double newValue = getNewValue(e, nom, den, monitor);
        nominator += func.f(e.x, newValue) - func.f(e.x, e.w);
        denominator += func.g(newValue) - func.g(e.w);
        e.w = newValue;
    }

    private double getNewValue(ExtendedElement e, double nom, double den, Monitor monitor) {
        assert init;
        //uwzglednic przypadek M==0 i m<0
        double m = FastMath.max(0, nom - func.f(e.x, e.w));
        double M = FastMath.max(0, den - func.g(e.w));
        monitor.computingUStarted();
        double newValue = func.u(lower, e.x, e.wl, e.wu, m, M, accuracy);
        monitor.computingUFinished();
        assert newValue <= e.wu && newValue >= e.wl : newValue + " | " + e.x + " " + e.wl + " " + e.wu + " " + m + " " + M;
        return newValue;
    }

    public double getNominator() {
        assert init;
        if (nominator < 0) {
            nominator = 0.0;
            for (ExtendedElement e : elements) {
                nominator += func.f(e.x, e.w);
            }
        }
        return nominator;
    }

    public double getNominator(double add) {
        assert init;
        double nomi = 0.0;
        for (ExtendedElement e : elements) {
            nomi += func.f(e.x, e.w + add);
        }
        return nomi;
    }

    public double getDenominator() {
        assert init;
        if (denominator < 0) {
            denominator = 0.0;
            for (ExtendedElement e : elements) {
                denominator += func.g(e.w);
            }
        }
        return denominator;
    }

    public double getDenominator(double add) {
        assert init;
        double denomi = 0.0;
        for (ExtendedElement e : elements) {
            denomi += func.g(e.w + add);
        }
        return denomi;
    }

    public double getY() {
        assert init;
        double nom = getNominator();
        double den = getDenominator();
        if (den == 0.0) {
            nom = getNominator(1e-10);
            den = getDenominator(1e-10);
        }
        double y = nom / den;
//        if(lower){
//            if(y<elements.get(0).p){
//                //mamy przypadek y<p
//                System.out.printf("przypadek lower: y=%.10f elem=%s func=%s (%.3f)\n",y,elements.get(0).toString(),func.getName(),func.getParam());
//                assert DoubleMath.fuzzyCompare(y, elements.get(0).p, 1e-6)>0;
//                y=elements.get(0).p;
//            }
//        }else{
//            if(y>elements.get(0).p){
//                //mamy przypadek y>p
//                System.out.printf("przypadek upper: y=%.10f elem=%s func=%s (%.3f)\n",y,elements.get(0).toString(),func.getName(),func.getParam());
//                assert DoubleMath.fuzzyCompare(y, elements.get(0).p, 1e-6)<0;
//                y=elements.get(0).p;
//            }
//        }
        return y;
    }

    private static class ExtendedElement extends Element {

        double w;

        public ExtendedElement(double p, double x, double wl, double wu) {
            super(p, x, wl, wu);
            w = wl;
        }

        @Override
        public String toString() {
            return "EE{" + "p=" + p + ", x=" + x + ", wl=" + wl + ", wu=" + wu + ", w=" + w + '}';
        }

    }
}
