package pl.edu.amu.wmi.min.relativecardinality.functions;

import static com.google.common.math.DoubleMath.fuzzyCompare;
import static com.google.common.math.DoubleMath.fuzzyEquals;
import org.apache.commons.math3.analysis.UnivariateFunction;
import org.apache.commons.math3.optim.MaxEval;
import org.apache.commons.math3.optim.nonlinear.scalar.GoalType;
import org.apache.commons.math3.optim.univariate.BrentOptimizer;
import org.apache.commons.math3.optim.univariate.MultiStartUnivariateOptimizer;
import org.apache.commons.math3.optim.univariate.SearchInterval;
import org.apache.commons.math3.optim.univariate.UnivariateObjectiveFunction;
import org.apache.commons.math3.optim.univariate.UnivariatePointValuePair;
import org.apache.commons.math3.random.RandomGenerator;
import org.apache.commons.math3.random.Well19937c;
import pl.edu.amu.wmi.min.relativecardinality.algorithm.Algorithm.Accuracy;

/**
 *
 * @author bikol
 */
abstract class AbstractFunctions implements Functions {

    private static final double D = 1e-8;
    private static final double[] DS = new double[]{1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 1e-10};
    protected static final RandomGenerator rand = new Well19937c(1536);

    private final String name;

    public AbstractFunctions(String name) {
        this.name = name;
    }

    @Override
    public double getParam() {
        return Double.NaN;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public double g(double a) {
        return a;
    }

    private double doP(double x, double u, double l) {
        assert l < u;
        double fu = g(f(x, u));
        double fl = g(f(x, l));
        double ff = fu - fl;
        double gg = g(u) - g(l);
        double fg = ff / gg;
        return Math.min(1.0, fg);
//        return (g(f(u, x)) - g(f(l, x))) / g(u) - g(l);
    }

    @Override
    public double p(boolean lower, double x, double l, double u) {
        double ld;
        if (l + D == l) {
            ld = Math.nextUp(l);
        } else {
            ld = l + D;
        }
        double toReturn = doP(x, u - l >= D ? u : ld, l);
        for (double d : DS) {
            if (l + d == l) {
                ld = Math.nextUp(l);
            } else {
                ld = l + d;
            }
            double c1 = doP(x, ld, l);
            if ((lower && toReturn > c1) || (!lower && toReturn < c1)) {
                toReturn = c1;
            }
        }
        return toReturn;
    }

    @Override
    public double u(final boolean lower, final double x, double l, double u, final double m, final double M, Accuracy accuracy) {
        if (l == u) {
            return l;
        } else {
            MultiStartUnivariateOptimizer opt = new MultiStartUnivariateOptimizer(new BrentOptimizer(1e-10, 1e-14), accuracy == Accuracy.STANDARD ? 3 : 1000, rand);
            UnivariateFunction func = new UnivariateFunction() {

                @Override
                public double value(double w) {
                    if ((g(w) + M) == 0.0) {
                        return (g(f(x, w + 1e-10)) + m) / (g(w + 1e-10) + M);
                    } else {
                        return (g(f(x, w)) + m) / (g(w) + M);
                    }
                }
            };
            UnivariatePointValuePair p = opt.optimize(
                    new MaxEval(accuracy == Accuracy.STANDARD ? 1000 : 1000000),
                    new SearchInterval(l, u),
                    lower ? GoalType.MINIMIZE : GoalType.MAXIMIZE,
                    new UnivariateObjectiveFunction(func));
            assert p.getPoint() >= l && p.getPoint() <= u;
//            System.out.println("Optimal val: "+p.getPoint()+" "+p.getValue());
            double lo = func.value(l);
            double up = func.value(u);
            //obsluzyc NaN
            if (Double.isNaN(lo)) {
                lo = lower ? 100 : -100;
            }
            if (Double.isNaN(up)) {
                up = lower ? 100 : -100;
            }
            if (lower) {
                if (fuzzyCompare(lo, p.getValue(), 1e-8) <= 0
                        || fuzzyCompare(up, p.getValue(), 1e-8) <= 0) {
                    if (fuzzyCompare(lo, up, 1e-8) < 0) {
                        return fuzzyEquals(lo, p.getValue(), 1e-8) ? p.getPoint() : l;
//                        return l;
                    } else {
                        return u;
                    }
                } else {
                    return p.getPoint();
                }
            } else {
                if (fuzzyCompare(lo, p.getValue(), 1e-8) >= 0 || fuzzyCompare(up, p.getValue(), 1e-8) >= 0) {
                    if (fuzzyCompare(lo, up, 1e-8) > 0) {
                        return fuzzyEquals(lo, p.getValue(), 1e-8) ? p.getPoint() : l;
//                        return l;
                    } else {
                        return u;
                    }
                } else {
                    return p.getPoint();
                }
            }
        }
    }

}
